@@ -2,12 +2,13 @@ from typing import TYPE_CHECKING, List, Type, Union
 
 from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
 from sqlalchemy.engine import Engine
-from sqlalchemy.ext.asyncio import AsyncEngine
+from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
+from sqlalchemy.orm import Session, sessionmaker
 from starlette.applications import Starlette
 from starlette.exceptions import HTTPException
 from starlette.requests import Request
 from starlette.responses import RedirectResponse, Response
-from starlette.routing import Mount, Route, Router
+from starlette.routing import Mount, Route
 from starlette.staticfiles import StaticFiles
 from starlette.templating import Jinja2Templates
 
@@ -32,11 +33,13 @@ class BaseAdmin:
         app: Starlette,
         engine: Union[Engine, AsyncEngine],
         base_url: str = "/admin",
+        title: str = "Admin",
+        logo_url: str = None,
     ) -> None:
         self.app = app
         self.engine = engine
         self.base_url = base_url
-        self._model_admins: List[Type["ModelAdmin"]] = []
+        self._model_admins: List["ModelAdmin"] = []
 
         self.templates = Jinja2Templates("templates")
         self.templates.env.loader = ChoiceLoader(
@@ -46,9 +49,12 @@ class BaseAdmin:
             ]
         )
         self.templates.env.globals["min"] = min
+        self.templates.env.globals["admin_title"] = title
+        self.templates.env.globals["admin_logo_url"] = logo_url
+        self.templates.env.globals["model_admins"] = self.model_admins
 
     @property
-    def model_admins(self) -> List[Type["ModelAdmin"]]:
+    def model_admins(self) -> List["ModelAdmin"]:
         """Get list of ModelAdmins lazily.
 
         Returns:
@@ -57,21 +63,13 @@ class BaseAdmin:
 
         return self._model_admins
 
-    def _find_model_admin(self, identity: str) -> Type["ModelAdmin"]:
+    def _find_model_admin(self, identity: str) -> "ModelAdmin":
         for model_admin in self.model_admins:
             if model_admin.identity == identity:
                 return model_admin
 
         raise HTTPException(status_code=404)
 
-    def _not_found_response(self, request: Request) -> Response:
-        context = {"request": request, "status_code": 404, "message": "Not found."}
-        return self.templates.TemplateResponse("error.html", context, status_code=404)
-
-    def _unathorized_response(self, request: Request) -> Response:
-        context = {"request": request, "status_code": 401, "message": "Unauthorized."}
-        return self.templates.TemplateResponse("error.html", context, status_code=401)
-
     def register_model(self, model: Type["ModelAdmin"]) -> None:
         """Register ModelAdmin to the Admin.
 
@@ -91,12 +89,44 @@ class BaseAdmin:
 
         # Set database engine from Admin instance
         model.engine = self.engine
-        model.sessionmaker = model._get_sessionmaker(model.engine)
+        if isinstance(model.engine, Engine):
+            model.sessionmaker = sessionmaker(bind=model.engine, class_=Session)
+            model.async_engine = False
+        else:
+            model.sessionmaker = sessionmaker(bind=model.engine, class_=AsyncSession)
+            model.async_engine = True
 
-        self._model_admins.append(model)
+        self._model_admins.append((model()))
 
 
-class Admin(BaseAdmin):
+class BaseAdminView(BaseAdmin):
+    async def _list(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+
+    async def _create(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.can_create or not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+
+    async def _details(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.can_view_details or not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+
+    async def _delete(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.can_delete or not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+
+    async def _edit(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.can_edit or not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+
+
+class Admin(BaseAdminView):
     """Main entrypoint to admin interface.
 
     ???+ usage
@@ -124,25 +154,44 @@ class Admin(BaseAdmin):
         app: Starlette,
         engine: Union[Engine, AsyncEngine],
         base_url: str = "/admin",
+        title: str = "Admin",
+        logo_url: str = None,
     ) -> None:
         """
         Args:
             app: Starlette or FastAPI application.
             engine: SQLAlchemy engine instance.
             base_url: Base URL for Admin interface.
+            title: Admin title.
+            logo_url: URL of logo to be displayed instead of title.
         """
 
         assert isinstance(engine, (Engine, AsyncEngine))
-        super().__init__(app=app, engine=engine, base_url=base_url)
+        super().__init__(
+            app=app, engine=engine, base_url=base_url, title=title, logo_url=logo_url
+        )
 
         statics = StaticFiles(packages=["sqladmin"])
 
-        router = Router(
+        def http_exception(request: Request, exc: Exception) -> Response:
+            assert isinstance(exc, HTTPException)
+            context = {
+                "request": request,
+                "status_code": exc.status_code,
+                "message": exc.detail,
+            }
+            return self.templates.TemplateResponse(
+                "error.html", context, status_code=exc.status_code
+            )
+
+        admin = Starlette(
             routes=[
                 Mount("/statics", app=statics, name="statics"),
                 Route("/", endpoint=self.index, name="index"),
                 Route("/{identity}/list", endpoint=self.list, name="list"),
-                Route("/{identity}/detail/{pk}", endpoint=self.detail, name="detail"),
+                Route(
+                    "/{identity}/details/{pk}", endpoint=self.details, name="details"
+                ),
                 Route(
                     "/{identity}/delete/{pk}",
                     endpoint=self.delete,
@@ -155,11 +204,16 @@ class Admin(BaseAdmin):
                     name="create",
                     methods=["GET", "POST"],
                 ),
-            ]
+                Route(
+                    "/{identity}/edit/{pk}",
+                    endpoint=self.edit,
+                    name="edit",
+                    methods=["GET", "POST"],
+                ),
+            ],
+            exception_handlers={HTTPException: http_exception},
         )
-        self.app.mount(base_url, app=router, name="admin")
-
-        self.templates.env.globals["model_admins"] = self.model_admins
+        self.app.mount(base_url, app=admin, name="admin")
 
     async def index(self, request: Request) -> Response:
         """Index route which can be overriden to create dashboards."""
@@ -169,20 +223,16 @@ class Admin(BaseAdmin):
     async def list(self, request: Request) -> Response:
         """List route to display paginated Model instances."""
 
+        await self._list(request)
+
         model_admin = self._find_model_admin(request.path_params["identity"])
 
         page = int(request.query_params.get("page", 1))
         page_size = int(request.query_params.get("page_size", 0))
+        search = request.query_params.get("search", None)
 
-        pagination = await model_admin.list(page, page_size)
-
-        if pagination.page != 1:
-            url = str(request.url.include_query_params(page=page - 1))
-            pagination.previous_page_url = url
-
-        if (pagination.page * pagination.page_size) < pagination.count:
-            url = str(request.url.include_query_params(page=page + 1))
-            pagination.next_page_url = url
+        pagination = await model_admin.list(page, page_size, search)
+        pagination.add_pagination_urls(request.url)
 
         context = {
             "request": request,
@@ -190,18 +240,18 @@ class Admin(BaseAdmin):
             "pagination": pagination,
         }
 
-        return self.templates.TemplateResponse("list.html", context)
+        return self.templates.TemplateResponse(model_admin.list_template, context)
 
-    async def detail(self, request: Request) -> Response:
-        """Detail route."""
+    async def details(self, request: Request) -> Response:
+        """Details route."""
+
+        await self._details(request)
 
         model_admin = self._find_model_admin(request.path_params["identity"])
-        if not model_admin.can_view_details:
-            return self._unathorized_response(request)
 
         model = await model_admin.get_model_by_pk(request.path_params["pk"])
         if not model:
-            return self._not_found_response(request)
+            raise HTTPException(status_code=404)
 
         context = {
             "request": request,
@@ -210,19 +260,19 @@ class Admin(BaseAdmin):
             "title": model_admin.name,
         }
 
-        return self.templates.TemplateResponse("detail.html", context)
+        return self.templates.TemplateResponse(model_admin.details_template, context)
 
     async def delete(self, request: Request) -> Response:
         """Delete route."""
 
+        await self._delete(request)
+
         identity = request.path_params["identity"]
         model_admin = self._find_model_admin(identity)
-        if not model_admin.can_delete:
-            return self._unathorized_response(request)
 
         model = await model_admin.get_model_by_pk(request.path_params["pk"])
         if not model:
-            return self._not_found_response(request)
+            raise HTTPException(status_code=404)
 
         await model_admin.delete_model(model)
 
@@ -231,10 +281,10 @@ class Admin(BaseAdmin):
     async def create(self, request: Request) -> Response:
         """Create model endpoint."""
 
+        await self._create(request)
+
         identity = request.path_params["identity"]
         model_admin = self._find_model_admin(identity)
-        if not model_admin.can_create:
-            return self._unathorized_response(request)
 
         Form = await model_admin.scaffold_form()
         form = Form(await request.form())
@@ -246,11 +296,11 @@ class Admin(BaseAdmin):
         }
 
         if request.method == "GET":
-            return self.templates.TemplateResponse("create.html", context)
+            return self.templates.TemplateResponse(model_admin.create_template, context)
 
         if not form.validate():
             return self.templates.TemplateResponse(
-                "create.html",
+                model_admin.create_template,
                 context,
                 status_code=400,
             )
@@ -262,3 +312,40 @@ class Admin(BaseAdmin):
             request.url_for("admin:list", identity=identity),
             status_code=302,
         )
+
+    async def edit(self, request: Request) -> Response:
+        """Edit model endpoint."""
+
+        await self._edit(request)
+
+        identity = request.path_params["identity"]
+        model_admin = self._find_model_admin(identity)
+
+        model = await model_admin.get_model_by_pk(request.path_params["pk"])
+        if not model:
+            raise HTTPException(status_code=404)
+
+        Form = await model_admin.scaffold_form()
+        context = {
+            "request": request,
+            "model_admin": model_admin,
+        }
+
+        if request.method == "GET":
+            context["form"] = Form(obj=model)
+            return self.templates.TemplateResponse(model_admin.edit_template, context)
+
+        form = Form(await request.form())
+        if not form.validate():
+            return self.templates.TemplateResponse(
+                model_admin.edit_template,
+                context,
+                status_code=400,
+            )
+
+        await model_admin.update_model(pk=request.path_params["pk"], data=form.data)
+
+        return RedirectResponse(
+            request.url_for("admin:list", identity=identity),
+            status_code=302,
+        )
